import asyncio
import os
from dataclasses import dataclass
from datetime import datetime, timedelta

import pytest
from aiohttp import web
from aiohttp.abc import Application, Request
from aiohttp.test_utils import TestServer
from aiohttp.web_response import json_response
from httpx import AsyncClient


async def create(request: Request):
    now = datetime.now()
    host = request.headers['host']
    request.app['sites'].append(request.headers['authorisation'])
    return json_response(
        {
            'message': 'New site created successfully',
            'secret_key': 'testing',
            'site_creation': now.isoformat(),
            'site_expiration': (now + timedelta(days=365)).isoformat(),
            'sites_created_24h': 0,
            'upload_expiration': (now + timedelta(hours=1)).isoformat(),
            'url': f'http://{host}/testing-site/',
        }
    )


async def upload(request: Request):
    body = await request.read()
    request.app['files'][request.path] = {'body': body, 'content-type': request.headers.get('content-type')}
    return json_response({'size': 123, 'total_site_size': 1234})


async def commit_update(request: Request):
    obj = await request.json()
    request.app['statuses'][request.match_info['path']] = obj
    return json_response({}, status=201)


routes = [
    web.post('/create/', create),
    web.post('/testing-site/{file:.*}', upload),
    web.post('/github/{path:.*}', commit_update),
]


@pytest.fixture(name='loop')
def _fix_loop():
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    return loop


@pytest.fixture(name='await_')
def _fix_await(loop):
    return loop.run_until_complete


@dataclass
class DummyServer:
    server: TestServer
    app: Application
    server_name: str

    async def stop(self):
        await self.server.close()

    @classmethod
    async def create(cls, loop, app_context) -> 'DummyServer':
        app = web.Application()
        app.add_routes(routes)
        app.update(app_context)

        server = TestServer(app)
        await server.start_server(loop=loop)

        return cls(server, app, f'http://localhost:{server.port}')


@pytest.fixture(name='dummy_server')
def _fix_dummy_server(loop):
    ctx = {'sites': [], 'files': {}, 'statuses': {}}
    ds = loop.run_until_complete(DummyServer.create(loop, ctx))

    yield ds

    loop.run_until_complete(ds.stop())


class SetEnv:
    def __init__(self):
        self.env_defaults = {}

    def set(self, name, value):
        self.env_defaults[name] = os.environ.get(name)
        os.environ[name] = value

    def clear(self):
        for name, value in self.env_defaults.items():
            if value is None:
                os.environ.pop(name)
            else:
                os.environ[name] = value


@pytest.fixture(name='env')
def _fix_env():
    setenv = SetEnv()

    yield setenv

    setenv.clear()


@pytest.fixture(name='async_client')
def _fix_async_client(await_):
    client = AsyncClient()
    await_(client.__aenter__())

    yield client

    await_(client.__aexit__())



import pytest

from smokeshow.main import get_github_status_info, set_github_commit_status

from .conftest import DummyServer, SetEnv


def test_status_success(tmp_path):
    f = tmp_path / 'index.html'
    f.write_text('<h1>testing</h1>\n<span class="pc_cov">55.44%</span>')
    assert get_github_status_info(tmp_path, 'test {coverage-percentage}', 40) == ('success', 'test 55.44%')


@pytest.mark.parametrize(
    'coverage,description,threshold,output',
    [
        ('55.4', 'test {coverage-percentage}', 40, ('success', 'test 55.40%')),
        ('55.4', 'test {coverage-percentage}', None, ('success', 'test 55.40%')),
        ('55.4', 'test', None, ('success', 'test')),
        ('55', 'test {coverage-percentage}', 40, ('success', 'test 55.00%')),
        ('55.123456', 'test {coverage-percentage}', 40, ('success', 'test 55.12%')),
        ('55.129', 'test {coverage-percentage}', 40, ('success', 'test 55.13%')),
        ('55.4', 'test', 40, ('success', 'test')),
        ('55.4', '{COVERAGE-PERCENTAGE}', 60, ('failure', '55.40% < 60.00%')),
        ('55.4', '', 60, ('failure', '')),
        ('100', 'test {coverage-percentage}', 100, ('success', 'test 100.00%')),
        ('99.99', 'test {coverage-percentage}', 100, ('failure', 'test 99.99% < 100.00%')),
    ],
)
def test_statues(tmp_path, coverage, description, threshold, output):
    f = tmp_path / 'index.html'
    f.write_text(f'<h1>testing</h1>\n<span class="pc_cov">{coverage}%</span>')
    assert get_github_status_info(tmp_path, description, threshold) == output


def test_no_index(tmp_path):
    f = tmp_path / 'foobar.html'
    f.write_text('<h1>testing</h1>\n<span class="pc_cov">55.44%</span>')
    assert get_github_status_info(tmp_path, 'test {coverage-percentage}', 1) == ('success', 'test {COVERAGE NOT FOUND}')


def test_root_is_file(tmp_path):
    f = tmp_path / 'foobar.html'
    f.write_text('hello')
    assert get_github_status_info(f, 'test {coverage-percentage}', 50) == ('success', 'test {COVERAGE NOT FOUND}')


def test_coverage_not_found(tmp_path):
    f = tmp_path / 'index.html'
    f.write_text('<h1>testing</h1>')
    assert get_github_status_info(tmp_path, 'test {coverage-percentage}', 1) == ('success', 'test {COVERAGE NOT FOUND}')


def test_set_status(env: SetEnv, mocker, dummy_server: DummyServer, await_, async_client):
    mocker.patch('smokeshow.main.GITHUB_API_ROOT', dummy_server.server_name + '/github')

    env.set('GITHUB_REPOSITORY', 'samuelcolvin/foobar')
    env.set('SMOKESHOW_GITHUB_PR_HEAD_SHA', 'abc1234')
    env.set('SMOKESHOW_GITHUB_TOKEN', 'xxx')

    await_(set_github_commit_status(async_client, 'https://www.example.com/testing', 'success', 'testing'))

    assert dummy_server.app['statuses'] == {
        'repos/samuelcolvin/foobar/statuses/abc1234': {
            'state': 'success',
            'target_url': 'https://www.example.com/testing',
            'description': 'testing',
            'context': 'smokeshow',
        },
    }



import re

import httpx
import pytest

from smokeshow import upload
from smokeshow.main import fmt_size

from .conftest import DummyServer


def test_upload_file(tmp_path, dummy_server: DummyServer, await_):
    f = tmp_path / 'test.html'
    f.write_text('<h1>testing</h1>')
    url = await_(upload(f, auth_key='testing-auth-key', root_url=dummy_server.server_name))
    assert re.fullmatch(r'http://localhost:\d+/testing-site/', url)

    assert dummy_server.app['sites'] == ['testing-auth-key']
    assert dummy_server.app['files'] == {
        '/testing-site/test.html': {'body': b'<h1>testing</h1>', 'content-type': 'text/html'}
    }


def test_upload_dir(tmp_path, dummy_server: DummyServer, await_):
    (tmp_path / 'index.html').write_text('<h1>testing</h1>')
    (tmp_path / 'foo.js.map').write_text('{"x": 123}')
    (tmp_path / 'bar.unknown_extension').write_text('xxx')

    url = await_(upload(tmp_path, auth_key='testing-auth-key', root_url=dummy_server.server_name))
    assert re.fullmatch(r'http://localhost:\d+/testing-site/', url)

    assert dummy_server.app['sites'] == ['testing-auth-key']
    assert dummy_server.app['files'] == {
        '/testing-site/index.html': {
            'body': b'<h1>testing</h1>',
            'content-type': 'text/html',
        },
        '/testing-site/foo.js.map': {
            'body': b'{"x": 123}',
            'content-type': 'application/json',
        },
        '/testing-site/bar.unknown_extension': {
            'body': b'xxx',
            'content-type': None,
        },
    }


def test_upload_dir_error(tmp_path, dummy_server: DummyServer, await_, mocker):
    mocker_upload = mocker.patch(
        'smokeshow.main._upload_file', side_effect=ValueError('intentional error testing upload')
    )
    (tmp_path / 'index.html').write_text('<h1>testing</h1>')
    (tmp_path / 'foo.js.map').write_text('{"x": 123}')
    (tmp_path / 'bar.unknown_extension').write_text('xxx')
    with pytest.raises(ValueError):
        await_(upload(tmp_path, auth_key='testing-auth-key', root_url=dummy_server.server_name))
    mocker_upload.call_count == 3


def test_upload_file_http_error(tmp_path, dummy_server: DummyServer, await_, mocker):
    mocker_upload = mocker.patch.object(
        httpx.AsyncClient, 'post', side_effect=httpx.HTTPError('intentional error testing upload')
    )
    (tmp_path / 'index.html').write_text('<h1>testing</h1>')
    (tmp_path / 'foo.js.map').write_text('{"x": 123}')
    (tmp_path / 'bar.unknown_extension').write_text('xxx')
    with pytest.raises(ValueError):
        await_(upload(tmp_path, auth_key='testing-auth-key', root_url=dummy_server.server_name))
    mocker_upload.assert_called_once()


def test_upload_generate(mocker, tmp_path, dummy_server: DummyServer, await_):
    mocker_generate_key = mocker.patch('smokeshow.main.generate_key', return_value='mocked-generate-key')

    f = tmp_path / 'test.html'
    f.write_text('<h1>testing</h1>')
    url = await_(upload(f, root_url=dummy_server.server_name))
    assert re.fullmatch(r'http://localhost:\d+/testing-site/', url)

    assert dummy_server.app['sites'] == ['mocked-generate-key']
    assert dummy_server.app['files'] == {
        '/testing-site/test.html': {'body': b'<h1>testing</h1>', 'content-type': 'text/html'}
    }
    mocker_generate_key.assert_called_once()


@pytest.mark.parametrize(
    'number,pretty',
    [
        (123, '123B'),
        (4.2 * 1024, '4.2KB'),
        (4.2 * 1024 * 1024, '4.2MB'),
    ],
)
def test_statues(number, pretty):
    assert fmt_size(number) == pretty



import re
import sys

import httpx
import pytest
from dirty_equals import IsStr
from typer.testing import CliRunner

from smokeshow.main import cli

runner = CliRunner()


def test_help():
    result = runner.invoke(cli, ['--help'])
    assert result.exit_code == 0
    assert result.stdout == IsStr(
        regex=r'.*Smokeshow CLI v\d\.[\d.]+, see https://smokeshow\.helpmanual\.io for.*', regex_flags=re.S
    )


def test_version():
    result = runner.invoke(cli, ['--version'])
    assert result.exit_code == 0
    assert result.stdout == IsStr(regex=r'Smokeshow v\d\.[\d.]+\n')


def test_generate_key(mocker):
    mocker.patch('smokeshow.main.KEY_HASH_THRESHOLD', 2**237)

    result = runner.invoke(cli, ['generate-key'])
    assert result.exit_code == 0
    assert 'Success! Key found after' in result.stdout
    assert '    SMOKESHOW_AUTH_KEY=' in result.stdout


@pytest.mark.skipif(sys.version_info < (3, 8), reason="Mock doesn't work well with async code in 3.7")
def test_upload_success(tmp_path, mocker):
    mocker.patch('smokeshow.main.upload')
    f = tmp_path / 'test.html'
    f.write_text('<h1>testing</h1>')

    result = runner.invoke(cli, ['upload', str(tmp_path)])
    assert result.exit_code == 0, result.stdout
    assert result.stdout == ''


def test_upload_error(tmp_path, mocker):
    mocker_upload = mocker.patch('smokeshow.main.upload', side_effect=ValueError('intentional error testing upload'))
    f = tmp_path / 'test.html'
    f.write_text('<h1>testing</h1>')

    result = runner.invoke(cli, ['upload', str(tmp_path)])
    assert result.exit_code == 1, result.stdout
    assert result.stdout == 'intentional error testing upload\n'

    mocker_upload.assert_called_once()


def test_upload_http_error(tmp_path, mocker):
    mocker_upload = mocker.patch.object(
        httpx.AsyncClient, 'post', side_effect=httpx.HTTPError('testing file upload failure')
    )
    f = tmp_path / 'test.html'
    f.write_text('<h1>testing</h1>')
    (tmp_path / 'foo.js.map').write_text('{"x": 123}')

    result = runner.invoke(cli, ['upload', str(tmp_path), '--auth-key', 'testing-auth-key'], catch_exceptions=False)
    assert result.exit_code == 1, result.stdout
    assert result.stdout == 'Error creating ephemeral site testing file upload failure\n'
    mocker_upload.assert_called_once()



import asyncio
import base64
import hashlib
import os
import re
import sys
from mimetypes import guess_type
from pathlib import Path
from typing import List, Optional, Tuple, Union, cast

from httpx import AsyncClient, AsyncHTTPTransport, HTTPError
from typer import Argument, Exit, Option, Typer

from .version import VERSION

__all__ = 'cli', 'upload'

USER_AGENT = f'smokeshow-cli-v{VERSION}'
KEY_HASH_THRESHOLD_POW = 234
KEY_HASH_THRESHOLD = 2**KEY_HASH_THRESHOLD_POW
ROOT_URL = 'https://smokeshow.helpmanual.io'
DEFAULT_TIMEOUT = 30  # seconds
UPLOAD_FILE_TIMEOUT = 300  # seconds
REQUEST_RETRIES = 3

cli = Typer(
    name='smokeshow', help=f'Smokeshow CLI v{VERSION}, see https://smokeshow.helpmanual.io for more information.'
)


def version_callback(value: bool) -> None:
    if value:
        print(f'Smokeshow v{VERSION}')
        raise Exit()


@cli.callback()
def _cli_callback(
    _version: bool = Option(None, '--version', callback=version_callback, is_eager=True),
) -> None:
    pass


@cli.command(help='Generate a new upload key')
def generate_key() -> str:
    print(
        'Searching for a key with valid hash '
        f'(the numeric representation of its sha-256 hash needs to be less than 2^{KEY_HASH_THRESHOLD_POW}). '
        'Hold tight, this might take a minute...'
    )
    attempts = 0
    while True:
        attempts += 1
        seed = os.urandom(50)
        h = int.from_bytes(hashlib.sha256(seed).digest(), 'big')
        if attempts % 100_000 == 0:
            print('.', end='', flush=True)
        if h < KEY_HASH_THRESHOLD:
            key = base64.b64encode(seed).decode().rstrip('=')
            print(f"\nSuccess! Key found after {attempts:,} attempts:\n\n    SMOKESHOW_AUTH_KEY='{key}'\n")
            return key


@cli.command(name='upload', help='Upload one or more files to create a new site')
def cli_upload(
    path: Path = Argument(..., exists=True, dir_okay=True, file_okay=True, readable=True, resolve_path=True),
    auth_key: Optional[str] = Option(None, envvar='SMOKESHOW_AUTH_KEY'),
    root_url: str = Option(ROOT_URL, envvar='SMOKESHOW_ROOT_URL'),
    github_status_description: Optional[str] = Option(None, envvar='SMOKESHOW_GITHUB_STATUS_DESCRIPTION'),
    github_coverage_threshold: Optional[float] = Option(None, envvar='SMOKESHOW_GITHUB_COVERAGE_THRESHOLD'),
) -> None:
    try:
        asyncio.run(
            upload(
                path,
                auth_key=auth_key,
                github_status_description=github_status_description,
                github_coverage_threshold=github_coverage_threshold,
                root_url=root_url,
            )
        )
    except ValueError as e:
        print(e, file=sys.stderr)
        raise Exit(1)


async def upload(
    root_path: Path,
    *,
    auth_key: Optional[str] = None,
    github_status_description: Optional[str] = None,
    github_coverage_threshold: Optional[float] = None,
    root_url: str = ROOT_URL,
) -> str:
    if auth_key is None:
        print('No auth key provided, generating one now...')
        auth_key_use = generate_key()
    else:
        auth_key_use = auth_key

    if not root_path.exists():
        raise ValueError(f'root path "{root_path}" does not exist')

    transport = AsyncHTTPTransport(retries=REQUEST_RETRIES)
    async with AsyncClient(timeout=DEFAULT_TIMEOUT, transport=transport) as client:
        try:
            r = await client.post(
                root_url + '/create/', headers={'Authorisation': auth_key_use, 'User-Agent': USER_AGENT}
            )
        except HTTPError as err:
            raise ValueError(f'Error creating ephemeral site {err}')
        if r.status_code != 200:
            raise ValueError(f'Error creating ephemeral site {r.status_code}, response:\n{r.text}')

        obj = r.json()
        secret_key: str = obj['secret_key']
        upload_root: str = obj['url']
        assert upload_root.endswith('/'), upload_root

        # useful when uploading to a dev endpoint where the worker returns the production host in request.url
        if not upload_root.startswith(root_url):
            upload_root = re.sub('^https?://[^/]+', root_url, upload_root)

        if root_path.is_dir():
            tasks = []
            for p in root_path.glob('**/*'):
                if not p.is_file():
                    continue
                task = asyncio.create_task(_upload_file(client, secret_key, upload_root, p, p.relative_to(root_path)))
                tasks.append(task)

            print(f'Site created with root {upload_root}\nuploading {len(tasks)} files...')
            # if an error occurs other tasks will still be executed which does not make much sense here so better
            # cancel them explicitly
            try:
                results = await asyncio.gather(*tasks)
            except ValueError:
                await _handle_tasks(tasks)
                raise
            total_size = max(results)
        else:
            # root_path is a file
            print(f'Site created with root {upload_root}\nuploading 1 file...')
            total_size = await _upload_file(client, secret_key, upload_root, root_path, root_path.name)

        print(f'upload complete ✓ site size {fmt_size(total_size)}')
        print('go to', upload_root)
        if github_status_description is not None:
            state, description = get_github_status_info(root_path, github_status_description, github_coverage_threshold)
            await set_github_commit_status(client, upload_root, state, description)

    return upload_root


async def _handle_tasks(tasks: 'List[asyncio.Task[int]]') -> None:
    """cancel all tasks and ignore all exceptions along the way"""
    for task in tasks:
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass
        except Exception:
            pass


async def _upload_file(
    client: AsyncClient, secret_key: str, upload_root: str, file_path: Path, rel_path: Union[Path, str]
) -> int:
    """
    Raises:
        ValueError:
            - If the connecting to the host fails or the post request fails in any other way.
            - If the post request yields any other response code than 200
    """
    url_path = str(rel_path)
    headers = {'Authorisation': secret_key, 'User-Agent': USER_AGENT}
    ct = get_content_type(url_path)
    if ct:
        headers['Content-Type'] = ct
    try:
        r2 = await client.post(
            upload_root + url_path, content=file_path.read_bytes(), headers=headers, timeout=UPLOAD_FILE_TIMEOUT
        )
    except HTTPError:
        print(f'    ERROR! Error uploading file {file_path}')
        raise ValueError(f'Uploading {url_path} failed due to a network error')

    if r2.status_code == 200:
        upload_info = r2.json()
        print(f'    {url_path} ct={ct} size={fmt_size(upload_info["size"])}')
        return cast(int, upload_info['total_site_size'])
    else:
        print(f'    ERROR! {url_path} status={r2.status_code} response={r2.text}')
        raise ValueError(f'invalid response from "{url_path}" status={r2.status_code} response={r2.text}')


def get_content_type(url: str) -> Optional[str]:
    if re.search(r'\.(js|css)\.map$', url):
        return 'application/json'
    else:
        return guess_type(url)[0]


KB = 1024
MB = KB**2


def fmt_size(num: int) -> str:
    if num < KB:
        return f'{num:0.0f}B'
    elif num < MB:
        return f'{num / KB:0.1f}KB'
    else:
        return f'{num / MB:0.1f}MB'


GITHUB_API_ROOT = 'https://api.github.com'


def get_github_status_info(path: Path, description: str, coverage_threshold: Optional[float]) -> Tuple[str, str]:
    state = 'success'
    if '{coverage-percentage}' not in description.lower() and coverage_threshold is None:
        return state, description

    cov_sub = '{COVERAGE NOT FOUND}'
    index_path = path / 'index.html'
    if index_path.is_file():
        m = re.search(r'<span\s+class="pc_cov">\s*([\d.]+)%\s*</span>', index_path.read_text())
        if m:
            coverage = float(m.group(1))
            if coverage_threshold is not None and coverage < coverage_threshold:
                state = 'failure'
                cov_sub = f'{coverage:0.2f}% < {coverage_threshold:0.2f}%'
            else:
                cov_sub = f'{coverage:0.2f}%'

    description = re.sub('{coverage-percentage}', cov_sub, description, flags=re.I)
    return state, description


async def set_github_commit_status(client: AsyncClient, target_url: str, state: str, description: str) -> None:
    github_repo = os.environ['GITHUB_REPOSITORY']
    github_sha = os.environ.get('SMOKESHOW_GITHUB_PR_HEAD_SHA') or os.environ['GITHUB_SHA']
    url = f'{GITHUB_API_ROOT}/repos/{github_repo}/statuses/{github_sha}'
    print(f'setting status on github.com/{github_repo}#{github_sha:.7}, {state}: "{description}"')

    github_token = os.environ['SMOKESHOW_GITHUB_TOKEN']
    context = 'smokeshow'
    github_context = os.environ.get('SMOKESHOW_GITHUB_CONTEXT')
    if github_context:
        context += f' / {github_context}'
    r = await client.post(
        url,
        headers={'authorization': f'Bearer {github_token}', 'accept': 'application/vnd.github.v3+json'},
        json={
            'state': state,
            'target_url': target_url,
            'description': description,
            'context': context,
        },
    )
    if r.status_code != 201:
        raise ValueError(f'invalid response from "{url}" status={r.status_code} response={r.text}')
